"""Caching functions for the Github Auth module."""

import hashlib
import pickle  # noqa: S403
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, TypeVar

from cachetools import TTLCache
from dogpile.cache import CacheRegion, make_region, register_backend
from dogpile.cache.api import NO_VALUE, BackendFormatted, BackendSetType, CacheBackend, CachedValue
from dogpile.cache.util import compat

R = TypeVar('R')


def get_cache_region():
    return make_region(function_key_generator=cache_key_generator)


# This is a copy of the default key generator from dogpile, but
# with additional hashing to avoid using sensitive values as cache keys
def cache_key_generator(namespace: Optional[str], fn: Callable[..., Any], to_str: Callable[[object], str] = str):
    if namespace is None:
        namespace = f'{fn.__module__}:{fn.__name__}'  # noqa: WPS609 - direct magic attribute usage
    else:
        namespace = f'{fn.__module__}:{fn.__name__}|{namespace}'  # noqa: WPS609 - direct magic attribute usage

    # Remove the `self` argument, since it'll always change
    fn_args = compat.inspect_getargspec(fn)
    has_self = fn_args[0] and (fn_args[0][0] in {'self', 'cls'})  # type: ignore

    def generate_key(*args: object, **kw: object) -> str:
        if kw:
            raise ValueError('The dogpile.cache default key creation function does not accept keyword arguments.')
        if has_self:
            args = args[1:]

        # Encode the args since they may be sensitive
        arg_key = hashlib.sha224(''.join(map(to_str, args)).encode('utf-8')).hexdigest()
        return f'{namespace}|{arg_key}'

    return generate_key


_default_cache_backend = 'memory'
# Expiration is dogpile's expiration TTL
_default_expiration = 300

_default_cache_size = 100

# The Cache TTL is the underlying backend TTL, which should
# be greater than dogpile's
# https://dogpilecache.sqlalchemy.org/en/latest/api.html#memcached-backends
_default_cache_ttl = _default_expiration * 1.5  # noqa: WPS432

# This gives us shortcuts to the actual modules
_backend_map = {
    _default_cache_backend: _default_cache_backend,
    'memcache': 'dogpile.cache.memcached',
}

# Default settings
_default_backend_args = {
    'memory': {'maxsize': _default_cache_size, 'ttl': _default_cache_ttl},
    'memcache': {'url': '127.0.0.1', 'distributed_lock': True},
}


def configure_cache_region(cache_region: CacheRegion, settings: Dict[str, Any], prefix: str):
    backend_key = f'{prefix}.backend'
    expiration_key = f'{prefix}.expiration'

    # Determine the backend
    backend = settings.get(backend_key, _default_cache_backend)
    expiration = int(settings.get(expiration_key, _default_expiration))

    # Find all the args that make sense for the backend
    backend_arg_prefix = f'{prefix}.{backend}.'
    backend_args = {
        k[len(backend_arg_prefix) :]: v for k, v in settings.items() if k.startswith(backend_arg_prefix)  # noqa: E203
    }

    resolved_args = {**_default_backend_args[backend], **backend_args}

    # Configure the cache region
    cache_region.configure(
        _backend_map[backend], expiration_time=expiration, arguments=resolved_args, replace_existing_backend=True,
    )


if TYPE_CHECKING:  # pragma: no cover
    StringIndexedTTLCache = TTLCache[str, CachedValue]
else:
    StringIndexedTTLCache = TTLCache


class TTLBackend(CacheBackend):
    _cache_path = 'cache_path'
    cache: StringIndexedTTLCache
    persisted_cache_path: Optional[str]

    def __init__(self, arguments: Dict[str, Any]):
        p_cache_path = arguments.pop(self._cache_path, None)

        if isinstance(p_cache_path, str):
            self.persisted_cache_path = p_cache_path
        else:
            self.persisted_cache_path = None

        # A potentially persisted cache for all items to keep in cache
        self.cache = TTLCache(**arguments)

        if self.persisted_cache_path:
            Path(self.persisted_cache_path).parent.mkdir(parents=True, exist_ok=True)

            try:
                # If we find a cache file and no argument was modified, then we retrieve the cache in file
                with open(self.persisted_cache_path, 'rb') as f:
                    pickled_cache = pickle.load(f)  # noqa: S301 - pickle usage
                    if all(getattr(self.cache, arg_key) == getattr(pickled_cache, arg_key) for arg_key in arguments.keys()):
                        self.cache = pickled_cache  # noqa: WPS220 - deep nesting

            except (FileNotFoundError, EOFError):
                pass

    def get(self, key: str) -> BackendFormatted:
        val = self.cache.get(key, NO_VALUE)
        if isinstance(val, NO_VALUE.__class__):
            return val
        return val

    def set(self, key: str, value: BackendSetType) -> None:  # noqa: WPS125, A003
        if not isinstance(value, CachedValue):  # pragma: no cover
            value = CachedValue(value, {})

        self.cache[key] = value
        if self.persisted_cache_path:
            self.persist_cache()

    def delete(self, key: str) -> None:
        sentinel = object()
        if self.cache.pop(key, sentinel) is not sentinel and self.persisted_cache_path:
            self.persist_cache()

    def persist_cache(self) -> None:
        assert self.persisted_cache_path
        with open(self.persisted_cache_path, 'wb') as f:
            pickle.dump(self.cache, f)


register_backend(_default_cache_backend, __name__, TTLBackend.__name__)



"""Transactions handler.

Usage examples:

- Default usage auto-applies all added operations.
```
with transaction() as ops:
    op = Operation(apply_fn=your_apply_fn, rollback_fn=your_rollback_fn)
    ops.add(op)
```

- We can disable auto-apply if necessary. In this case, the context manager
will check if some operations weren't applied at the end of the transaction.
```
op = None
with transaction(auto_apply=False) as ops:
    op = Operation(apply_fn=your_apply_fn, rollback_fn=your_rollback_fn)
    ops.add(create_user)
    op.apply()
```

"""

from __future__ import annotations

import abc
from contextlib import contextmanager
from contextvars import ContextVar
from enum import Enum
from types import TracebackType
from typing import Callable, Generic, Iterator, List, Optional, Type, TypeVar

txns = ContextVar('txns', default=[])


def reset_context():  # pragma: no cover
    # This method exposes a way to reset the ContextVar.
    txns.set([])


E = TypeVar('E', bound=Exception)


class TransactionManager:
    @property
    def transactions(self) -> List[Transaction]:
        # We store transactions of a hierarchie in a ContextVar to make sure
        # there are no conflicts in an async context.
        return txns.get()

    @property
    def current_txn(self) -> Optional[Transaction]:
        if self.transactions:
            # Return the current transaction.
            return self.transactions[-1]
        return None

    @current_txn.setter
    def current_txn(self, value: Transaction):
        self.transactions.append(value)

    def __enter__(self) -> Transaction:
        parent_transaction = self.current_txn
        self.current_txn = Transaction()
        if parent_transaction is not None:
            # We wrap the transaction to give it the same interface as a resource, and link it
            # to the parent transaction. That way, the parent transaction will be able to abort its
            # child transactions if any of them fails.
            parent_transaction.join(TransactionResourceManager(self.current_txn))
        return self.current_txn

    def __exit__(self, exc_type: Type[E], exc_value: E, traceback: TracebackType):
        assert self.current_txn
        if exc_value is None:
            self.current_txn.commit()
        else:
            self.current_txn.abort()
        self.transactions.pop()


class Status(Enum):
    active = 'Active'
    committed = 'Committed'
    commitfailed = 'Commit failed'


class Transaction:
    _resources: List[ResourceManager]

    def __init__(self):
        self.status = Status.active
        self._resources = []

    def join(self, resource: ResourceManager) -> None:
        assert self.status is Status.active  # noqa: S101
        self._resources.append(resource)

    def commit(self) -> None:
        assert self.status is Status.active  # noqa: S101

        try:
            self._commit_resources()
            self.status = Status.committed
        except Exception as e:  # pragma: no cover
            self.status = Status.commitfailed
            raise e

    def abort(self) -> None:
        exceptions: List[Exception] = []
        for rm in reversed(self._resources):
            try:
                rm.abort()
            except Exception as e:  # pragma: no cover
                exceptions.append(e)
        if exceptions:  # pragma: no cover
            raise Exception(exceptions)

    def _commit_resources(self) -> None:
        try:
            for rm in self._resources:
                rm.commit()
        except Exception:  # pragma: no cover
            # If an error occurs committing a transaction, we try
            # to revert the changes in each of the resource managers.
            for rm in reversed(self._resources):  # noqa: WPS440
                rm.abort()  # noqa: WPS441


class ResourceManager(abc.ABC):
    def __init__(self, txn: Transaction):
        self.txn = txn

    @abc.abstractmethod
    def abort(self) -> None:  # pragma: no cover
        ...

    @abc.abstractmethod
    def commit(self) -> None:  # pragma: no cover
        ...


@contextmanager
def transaction(auto_apply: bool = True) -> Iterator[OperationManager]:
    manager = TransactionManager()

    with manager as txn:
        ops = OperationManager(txn)
        txn.join(ops)

        yield ops

        if auto_apply:
            ops.apply()
        else:
            ops.ensure_applied()


class OperationState(Enum):
    pending = 'pending'
    applied = 'applied'
    errored = 'errored'
    reset = 'reset'


class InvalidState(Exception):
    ...


class OperationNotApplied(Exception):
    ...


T = TypeVar('T', covariant=True)

OperationApplyFn = Callable[[], Optional[T]]
OperationRollbackFn = Callable[[Optional[T]], None]


class Operation(Generic[T]):
    result: Optional[T]
    apply_fn: OperationApplyFn[T]
    rollback_fn: OperationRollbackFn[T]

    def __init__(self, apply_fn: OperationApplyFn[T], rollback_fn: OperationRollbackFn[T]):
        self.apply_fn = apply_fn
        self.rollback_fn = rollback_fn
        self.state = OperationState.pending
        self.result = None

    def apply(self) -> None:
        if self.state != OperationState.pending:
            raise InvalidState

        try:
            self.result = self.apply_fn()
            self.state = OperationState.applied
        except Exception as ex:
            self.state = OperationState.errored
            raise Exception('Failed to apply operation') from ex

    def rollback(self) -> None:
        if self.state != OperationState.applied:
            raise InvalidState

        self.rollback_fn(self.result)
        self.state = OperationState.reset


class OperationManager(ResourceManager):  # noqa: WPS214 - too many methods
    operations: List[Operation[object]]

    def __init__(self, txn: Transaction):
        super().__init__(txn)
        self.operations = []

    def add(self, operation: Operation[object]) -> None:
        self.operations.append(operation)

    def apply(self) -> None:
        for op in self.operations:
            op.apply()

    def abort(self) -> None:
        for op in reversed(self.operations):
            if op.state == OperationState.applied:
                op.rollback()

    def ensure_applied(self) -> None:
        for op in self.operations:
            if op.state != OperationState.applied:
                raise OperationNotApplied(op.state)

    def commit(self) -> None:
        pass


class TransactionResourceManager(ResourceManager):
    def abort(self) -> None:
        self.txn.abort()

    def commit(self) -> None:
        pass



"""Main entrypoint for the Jinja2 extension.

Responsable for adding all of the filters to the Jinja
Environment object.
"""

from jinja2 import Environment
from jinja2.ext import Extension
from outcome.utils.jinja2.regex_replace import regex_replace


class Extend(Extension):
    """Extends the Jinja environment with the custom filters."""

    def __init__(self, environment: Environment):
        """Initialize the plugin, but also add the filters to the environment.

        Args:
            environment: The Jinja2 environment object.
        """
        super()

        environment.filters['regex_replace'] = regex_replace



"""A utility to find the nearest pyproject.toml file, in the ancestry."""

from pathlib import Path
from typing import Optional, Union

pyproject_toml = 'pyproject.toml'


def find_pyproject_toml(starting_directory: Union[Path, str]) -> Optional[Path]:
    current_path = Path(starting_directory)

    if not current_path.is_dir():
        current_path = current_path.parent

    for path in (current_path, *current_path.parents):
        possible_pyproject_file = path / pyproject_toml
        if possible_pyproject_file.is_file():
            return possible_pyproject_file

    return None
